//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// TypeEnvironment.cc --
//
// Author           : Aaron B. Moss
// Created On       : Sun May 17 12:19:47 2015
// Last Modified By : Aaron B. Moss
// Last Modified On : Fri Jun 29 15:51:00 2018
// Update Count     : 5
//

#include <cassert>                     // for assert
#include <algorithm>                   // for copy, set_intersection
#include <iterator>                    // for ostream_iterator, insert_iterator
#include <list>                        // for list
#include <vector>                      // for vector
#include <unordered_map>               // for unordered_map
#include <utility>                     // for pair, move

#include "Common/InternedString.h"     // for interned_string
#include "Common/PassVisitor.h"        // for PassVisitor<GcTracer>
#include "Common/option.h"             // for option<T>
#include "Common/utility.h"            // for maybeClone
#include "SymTab/Indexer.h"            // for Indexer
#include "SynTree/GcTracer.h"          // for PassVisitor<GcTracer>
#include "SynTree/Type.h"              // for Type, FunctionType, Type::Fora...
#include "SynTree/TypeSubstitution.h"  // for TypeSubstitution
#include "Tuples/Tuples.h"             // for isTtype
#include "TypeEnvironment.h"
#include "typeops.h"                   // for occurs
#include "Unify.h"                     // for unifyInexact

namespace ResolvExpr {
	#if 0
	#define PRE_POST_VALIDATE auto dbg = ValidateGuard{this, __func__};
	#define PRE_POST_VALIDATE_NOM auto dbg = ValidateGuard{this};
	struct ValidateGuard {
		const TypeEnvironment* env;
		const TypeEnvironment::Classes* old_classes;
		const TypeEnvironment::Bindings* old_bindings;
		const char* last_fn;

		void validate_classes( const TypeEnvironment::Classes* c ) {
			typedef TypeEnvironment::Classes C;

			assertf( c != nullptr, "classes non-null" );
			
			C::Mode cmode = c->get_mode();
			if ( cmode == C::BASE ) {
				// xxx - consistency checks
			} else {
				assertf( cmode == C::ADD || cmode == C::REM || cmode == C::ADDTO || cmode == C::REMFROM, "valid classes mode" );
				validate_classes( c->get_base() );
			}
		}

		void validate_bindings( const TypeEnvironment::Bindings* b ) {
			typedef TypeEnvironment::Bindings B;

			assertf( b != nullptr, "bindings non-null" );

			B::Mode bmode = b->get_mode();
			if ( bmode == B::BASE ) {
				// xxx - consistency checks
			} else {
				assertf( bmode == B::REM || bmode == B::INS || bmode == B::UPD, "valid bindings mode" );
				validate_bindings( b->get_base() );
			}
		}

		void validate_env() {
			validate_classes( env->classes );
			validate_bindings( env->bindings );
			
			// xxx - joint validation
		}

		ValidateGuard(const TypeEnvironment* e)
			: env{e}, old_classes{e->classes}, old_bindings{e->bindings}, last_fn{e->last_fn}
			{ validate_env(); }
		ValidateGuard(const TypeEnvironment* e, const char* fn)
			: env{e}, old_classes{e->classes}, old_bindings{e->bindings}, last_fn{fn} 
			{ validate_env(); }
		~ValidateGuard() {
			validate_env();
			if ( env->classes != old_classes ) validate_classes( old_classes );
			if ( env->bindings != old_bindings ) validate_bindings( old_bindings );
			const_cast<TypeEnvironment*>(env)->last_fn = last_fn;
		}
	};
	#else
	#define PRE_POST_VALIDATE
	#define PRE_POST_VALIDATE_NOM
	#endif

	void printAssertionSet( const AssertionSet &assertions, std::ostream &os, int indent ) {
		for ( AssertionSet::const_iterator i = assertions.begin(); i != assertions.end(); ++i ) {
			i->first->print( os, indent );
			if ( i->second.isUsed ) {
				os << "(used)";
			} else {
				os << "(not used)";
			} // if
		} // for
	}

	void printOpenVarSet( const OpenVarSet &openVars, std::ostream &os, int indent ) {
		os << std::string( indent, ' ' );
		for ( OpenVarSet::const_iterator i = openVars.begin(); i != openVars.end(); ++i ) {
			os << i->first << "(" << i->second << ") ";
		} // for
	}

	std::pair<interned_string, interned_string> TypeEnvironment::mergeClasses( 
			interned_string root1, interned_string root2 ) {
		PRE_POST_VALIDATE
		// merge classes
		Classes* newClasses = classes->merge( root1, root2 );

		// determine new root
		assertf(classes->get_mode() == Classes::REMFROM, "classes updated to REMFROM by merge");
		auto ret = std::make_pair( classes->get_root(), classes->get_child() );
		
		// finalize classes
		classes = newClasses;
		return ret;
	}

	ClassRef TypeEnvironment::lookup( interned_string var ) const {
		interned_string root = classes->find_or_default( var, nullptr );
		if ( root ) return { this, root };
		else return { nullptr, var };
	}

	bool isFtype( Type *type ) {
		if ( dynamic_cast< FunctionType* >( type ) ) {
			return true;
		} else if ( TypeInstType *typeInst = dynamic_cast< TypeInstType* >( type ) ) {
			return typeInst->get_isFtype();
		} // if
		return false;
	}

	bool tyVarCompatible( const TypeDecl::Data & data, Type *type ) {
		switch ( data.kind ) {
		  case TypeDecl::Dtype:
			// to bind to an object type variable, the type must not be a function type.
			// if the type variable is specified to be a complete type then the incoming
			// type must also be complete
			// xxx - should this also check that type is not a tuple type and that it's not a ttype?
			return ! isFtype( type ) && (! data.isComplete || type->isComplete() );
		  case TypeDecl::Ftype:
			return isFtype( type );
		  case TypeDecl::Ttype:
			// ttype unifies with any tuple type
			return dynamic_cast< TupleType * >( type ) || Tuples::isTtype( type );
		} // switch
		return false;
	}

	bool TypeEnvironment::bindVar( TypeInstType* typeInst, Type* bindTo, 
			const TypeDecl::Data& data, AssertionSet& need, AssertionSet& have, 
			const OpenVarSet& openVars, WidenMode widenMode, const SymTab::Indexer& indexer ) {
		PRE_POST_VALIDATE
		// remove references from other, so that type variables can only bind to value types
		bindTo = bindTo->stripReferences();
		
		auto tyVar = openVars.find( typeInst->get_name() );
		assert( tyVar != openVars.end() );
		if ( ! tyVarCompatible( tyVar->second, bindTo ) ) return false;

		if ( occurs( bindTo, typeInst->get_name(), *this ) ) return false;

		if ( ClassRef curClass = lookup( typeInst->get_name() ) ) {
			BoundType curData = curClass.get_bound();
			if ( curData.type ) {
				Type* common = nullptr;
				// attempt to unify equivalence class type (which needs its qualifiers restored) 
				// with the target type
				Type* newType = curData.type->clone();
				newType->get_qualifiers() = typeInst->get_qualifiers();
				if ( unifyInexact( newType, bindTo, *this, need, have, openVars, 
						widenMode & WidenMode{ curData.allowWidening, true }, indexer, common ) ) {
					if ( common ) {
						// update bound type to common type
						common->get_qualifiers() = Type::Qualifiers{};
						curData.type = common;
						bindings = bindings->set( curClass.update_root(), curData );
					}
					return true;
				} else return false;
			} else {
				// update bound type to other type
				curData.type = bindTo->clone();
				curData.type->get_qualifiers() = Type::Qualifiers{};
				curData.allowWidening = widenMode.widenFirst && widenMode.widenSecond;
				bindings = bindings->set( curClass.get_root(), curData );
			}
		} else {
			// make new class consisting entirely of this variable
			BoundType curData{ 
				bindTo->clone(), widenMode.widenFirst && widenMode.widenSecond, data };
			curData.type->get_qualifiers() = Type::Qualifiers{};
			classes = classes->add( curClass.get_root() );
			bindings = bindings->set( curClass.get_root(), curData );
		}
		return true;
	}
	
	bool TypeEnvironment::bindVarToVar( TypeInstType* var1, TypeInstType* var2, 
			const TypeDecl::Data& data, AssertionSet& need, AssertionSet& have, 
			const OpenVarSet& openVars, WidenMode widenMode, const SymTab::Indexer& indexer ) {
		PRE_POST_VALIDATE
		ClassRef class1 = lookup( var1->get_name() );
		ClassRef class2 = lookup( var2->get_name() );
		
		// exit early if variables already bound together
		if ( class1 && class2 && class1 == class2 ) {
			BoundType data1 = class1.get_bound();
			// narrow the binding if needed
			if ( data1.allowWidening && widenMode.widenFirst != widenMode.widenSecond ) {
				data1.allowWidening = false;
				bindings = bindings->set( class1.get_root(), data1 );
			}
			return true;
		}

		BoundType data1 = class1 ? class1.get_bound() : BoundType{};
		BoundType data2 = class2 ? class2.get_bound() : BoundType{};

		bool widen1 = data1.allowWidening && widenMode.widenFirst;
		bool widen2 = data2.allowWidening && widenMode.widenSecond;

		if ( data1.type && data2.type ) {
			// attempt to unify bound classes
			Type* common = nullptr;
			if ( unifyInexact( data1.type->clone(), data2.type->clone(), *this, need, have, 
					openVars, WidenMode{ widen1, widen2 }, indexer, common ) ) {
				// merge type variables
				interned_string root = 
					mergeClasses( class1.update_root(), class2.update_root() ).first;
				// update bindings
				data1.allowWidening = widen1 && widen2;
				if ( common ) {
					common->get_qualifiers() = Type::Qualifiers{};
					data1.type = common;
				}
				bindings = bindings->set( root, data1 );
			} else return false;
		} else if ( class1 && class2 ) {
			// both classes exist, only one bound -- merge type variables
			auto merged = mergeClasses( class1.get_root(), class2.get_root() );
			// remove subordinate binding
			bindings = bindings->erase( merged.second );
			// update root binding (narrow widening as needed, or set new binding for changed root)
			if ( data1.type ) {
				if ( data1.allowWidening != widen1 ) {
					data1.allowWidening = widen1;
					bindings = bindings->set( merged.first, data1 );
				} else if ( merged.first == class2.get_root() ) {
					bindings = bindings->set( merged.first, data1 );
				}
			} else /* if ( data2.type ) */ {
				if ( data2.allowWidening != widen2 ) {
					data2.allowWidening = widen2;
					bindings = bindings->set( merged.first, data2 );
				} else if ( merged.first == class1.get_root() ) {
					bindings = bindings->set( merged.first, data2 );
				}
			}
		} else if ( class1 ) {
			// add unbound var2 to class1
			classes = classes->add( class2.get_root() );
			auto merged = mergeClasses( class1.get_root(), class2.get_root() );
			// update bindings (narrow as needed, or switch binding to root)
			if ( merged.first == class2.get_root() ) {
				data1.allowWidening = widen1;
				bindings = bindings->erase( merged.second )->set( merged.first, data1 );
			} else if ( data1.allowWidening != widen1 ) {
				bindings = bindings->set( merged.first, data1 );
			}
		} else if ( class2 ) {
			// add unbound var1 to class1
			classes = classes->add( class1.get_root() );
			auto merged = mergeClasses( class1.get_root(), class2.get_root() );
			// update bindings (narrow as needed, or switch binding to root)
			if ( merged.first == class1.get_root() ) {
				data2.allowWidening = widen2;
				bindings = bindings->erase( merged.second )->set( merged.first, data2 );
			} else if ( data2.allowWidening != widen2 ) {
				bindings = bindings->set( merged.first, data2 );
			}
		} else {
			// make new class with pair of unbound vars
			classes = classes->add( class1.get_root() )->add( class2.get_root() );
			auto merged = mergeClasses( class1.get_root(), class2.get_root() );
			bindings = bindings->set( merged.first, BoundType{ nullptr, widen1 && widen2, data } );
		}
		return true;
	}

	void TypeEnvironment::add( const Type::ForallList &tyDecls ) {
		PRE_POST_VALIDATE
		for ( Type::ForallList::const_iterator i = tyDecls.begin(); i != tyDecls.end(); ++i ) {
			interned_string name = (*i)->get_name();
			classes = classes->add( name );
			bindings = bindings->set( name, *i );
		} // for
	}

	void TypeEnvironment::add( const TypeSubstitution & sub ) {
		PRE_POST_VALIDATE
		interned_string not_found{nullptr};

		for ( auto p : sub ) {
			interned_string var = p.first;
			
			// filter overlapping classes out of existing environment
			// xxx - this is a very shady assumption, but has worked for a long time...
			interned_string root = classes->find_or_default( var, not_found );
			if ( root != not_found ) {
				classes = classes->remove_class( root );
				bindings = bindings->erase( root );
			}

			// Minimal assumptions. Not technically correct, but might be good enough, and
			// is the best we can do at the moment since information is lost in the
			// transition to TypeSubstitution
			TypeDecl::Data data{ TypeDecl::Dtype, false };

			// add variable to class and bindings
			classes = classes->add( var );
			bindings = bindings->set( var, BoundType{ p.second->clone(), false, data } );
		}
	}

	void TypeEnvironment::makeSubstitution( TypeSubstitution &sub ) const {
		PRE_POST_VALIDATE_NOM
		bindings->for_each([&]( interned_string root, const BoundType& bound ){
			classes->for_class(root, [&]( interned_string var ) {
				if ( bound.type ) {
					sub.add( var, bound.type );
				} else if ( var != root ) {
					sub.add( var, new TypeInstType{ 
						Type::Qualifiers{}, root, bound.data.kind == TypeDecl::Ftype } );
				}
			});
		});

		sub.normalize();
	}

	void TypeEnvironment::print( std::ostream &os, Indenter indent ) const {
		bindings->for_each([&]( interned_string root, const BoundType& bound ) {
			os << "( ";
			classes->for_class( root, [&os]( interned_string var ) { os << var << " "; } );
			os << ")";
			if ( bound.type ) {
				os << " -> ";
				bound.type->print( os, indent+1 );
			}
			if ( ! bound.allowWidening ) {
				os << " (no widening)";
			}
			os << std::endl;
		});
	}

	namespace {
		// temporary representation of an equivalence class
		struct EqvClass {
			std::vector<interned_string> vars;
			BoundType bound;

			EqvClass() = default;
			EqvClass(const BoundType& b) : bound{b} {}
		};
	}

	void TypeEnvironment::simpleCombine( const TypeEnvironment &o ) {
		PRE_POST_VALIDATE
		// check for same environment
		if ( classes == o.classes && bindings == o.bindings ) return;

		// read out equivalence classes (avoids conflicting reroots)
		std::vector<EqvClass> ecs;
		o.bindings->for_each( [&]( interned_string root, const BoundType& bound ) {
			ecs.emplace_back(bound);
			o.classes->for_class( root, [&]( interned_string var ) {
				ecs.back().vars.push_back( var );
			} );
		} );

		// read equivalence classes into self
		for ( EqvClass& ec : ecs ) {
			// merge vars
			interned_string new_root{nullptr};
			for ( interned_string var : ec.vars ) {
				Classes* new_classes = classes->add( var );
				if ( new_classes == classes ) { var = classes->find( var ); }
				else { classes = new_classes; }
				new_root = new_root ? mergeClasses( new_root, var ).first : var;
			}
			// set binding
			bindings = bindings->set( new_root, ec.bound );
		}
	}

	bool TypeEnvrionment::combine( const TypeEnvironment& o, const SymTab::Indexer& indexer ) {
		// short-circuit for empty cases
		if ( o.isEmpty() ) return true;
		if ( isEmpty() ) {
			*this = o;
			return true;
		}

		/// Edit item for path state
		struct Edit {
			Classes::Mode mode;   ///< Type of change to a key
			interned_string root; ///< New/Previous root, if addTo/remFrom node

			Edit(Classes::Mode m, interned_string r = nullptr) : mode{m}, root{r} {}
		};

		// track class changes
		classes->reroot();
		PRE_POST_VALIDATE

		using EditEntry = std::pair<interned_string, Edit>;
		using EditList = std::list<EditEntry>;
		using IndexedEdits = std::vector<EditList::iterator>;

		EditList edits;
		std::unordered_map<interned_string, IndexedEdits> editIndex;

		// trace path from other environment
		const Classes* oclasses = o.classes;
		Classes::Mode omode = oclasses->get_mode();
		while ( omode != Classes::BASE ) {
			interned_string key = oclasses->get_key();
			IndexedEdits& forKey = editIndex[ key ];

			if ( forKey.empty() ) {
				// newly seen key, mark op
				interned_string root = ( omode == Classes::ADD || omode == Classes::REM ) ?
					nullptr : oclasses->get_root();
				auto it = edits.emplace( edits.begin(), key, Edit{ omode, root } );
				forKey.push_back( move(it) );
			} else {
				auto next = forKey.back();  // edit next applied to this key
				Classes::Mode nmode = next->second.mode; // next action applied

				switch ( omode ) {
					case Classes::ADD: {
						switch ( nmode ) {
							case Classes::REM: {
								// later removal, no net change to classes
								edits.erase( next );
								forKey.pop_back();
							} break;
							case Classes::ADDTO: {
								// later merge, prefix with this addition
								auto it = edits.emplace( edits.begin(), key, Edit{ omode } );
								forKey.push_back( move(it) );
							} break;
							default: assertf( false, "inconsistent mode" );
						}
					} break;
					case Classes::REM: {
						// later addition, no net change to classes
						assertf( nmode == Classes::ADD, "inconsistent mode" );
						edits.erase( next );
						forKey.pop_back();
					} break;
					case Classes::ADDTO: {
						// later unmerged from same class, no net change to classes
						assume( nmode == Classes::REMFROM, "inconsistent mode" );
						assume( oclasses->get_root() == next->second.root, "inconsistent tree" );
						edits.erase( next );
						forKey.pop_back();
					} break;
					case Classes::REMFROM: {
						interned_string root = oclasses->get_root();
						switch ( nmode ) {
							case Classes::REM: {
								// later removal, prefix with this unmerge
								auto it = edits.emplace( edits.begin(), key, Edit{ omode, root } );
								forKey.push_back( move(it) );
							} break;
							case Classes::ADDTO: {
								if ( root == next->second.root ) {
									// later merged back into same class, no net change
									edits.erase( next );
									forKey.pop_back();
								} else {
									// later merged into different class, prefix with this unmerge
									auto it = edits.emplace( 
										edits.begin(), key, Edit{ omode, root } );
									forKey.push_back( move(it) );
								}
							} break;
							default: assertf( false, "inconsistent mode" );
						}
					} break;
					default: assertf( false, "invalid mode" );
				}
			}

			oclasses = oclasses->get_base();
			omode = oclasses->get_mode();
		}
		assertf( oclasses == classes, "classes must be versions of the same map" );

		/// Edit item for binding changes
		struct BEdit {
			Bindings::Mode mode;  ///< Type of change to a key
			BoundType val;        ///< Updated key value, if applicable

			BEdit(Bindings::Mode m) : mode{m}, val{} {}
			BEdit(Bindings::Mode m, const BoundType& b) : mode{m}, val{b} {}
		};

		// track binding merges
		bindings->reroot();
		std::unordered_map<interned_string, BEdit> bedits; // edits to base bindings

		// trace path from other environment
		const Bindings* bbindings = o.bindings;
		Bindings::Mode bmode = bbindings->get_mode();
		while ( bmode != Bindings::BASE ) {
			interned_string key = bbindings->get_key();
			auto it = bedits.find( key );

			if ( it == bedits.end() ) {
				// newly seen key, mark operation
				if ( bmode == Bindings::REM ) {
					bedits.emplace_back( it, key, BEdit{ bmode } );
				} else {
					bedits.emplace_back( it, key, BEdit{ bmode, bbindings->get_val() } );
				}
			} else {
				Bindings::Mode& smode = it->second.mode;
				switch ( bmode ) {
					case Bindings::REM: {
						// later insertion, change to update
						assertf( smode == Bindings::INS, "inconsistent mode" );
						smode = Bindings::UPD;
					} break;
					case Bindings::INS: {
						switch ( smode ) {
							case Bindings::REM: {
								// later removal, no net change to map
								bedits.erase( it );
							} break;
							case Bindings::UPD: {
								// later update collapses to insertion
								smode = Bindings::INS;
							} break;
							default: assertf( false, "inconsistent mode" );
						}
					} break;
					case Bindings::UPD: {
						// later removal or update overrides this update
						assertf( smode == Bindings::REM || smode == Bindings::UPD, "inconsistent mode" );
					} break;
					default: assertf( false, "invalid mode" );
				}
			}

			bbindings = bbindings->get_base();
			bmode = bbindings->get_mode();
		}
		assume( bbindings == bindings, "bindings must be versions of same map" );

		// merge typeclasses (always possible, can always merge all classes into one if the 
		// bindings unify)
		for ( const EditEntry& edit : edits ) {
			interned_string key = edit.first;
			const Edit& e = edit.second;

			switch ( e.mode ) {
				case Classes::ADD: {
					// add new type variable
					classes = classes->add( key );
					// add binding for variable
					auto bit = bedits.find( key );
					if ( bit != bedits.end() ) {
						// take final value if it is a root in its own map
						const BEdit& be = bit->second;
						assertf( be.mode == Bindings::INS, "inconsistent binding" );
						bindings = bindings->set( key, std::move(be.val) );
						// remove update from edit set so that it isn't re-checked
						bedits.erase( bit );
					} else {
						// otherwise just put in default binding
						bindings = bindings->set( key, BoundType{} );
					}
				} break;
				case Classes::REM: {
					// do not remove type variable (merging)
				} break;
				case Classes::ADDTO: {
					// if previous REMFROM on this key, or ADDTO on this root, may not match 
					// current state in `classes`
					interned_string key_root = classes->find( key );
					interned_string merge_root = classes->find( e.root );
					// skip case where classes already merged
					if ( key_root == merge_root ) break;
					// merge classes and remove binding for new child
					interned_string new_child = mergeClasses( key_root, merge_root ).second;
					if ( bindings->count( new_child ) ) {
						// add new insert edit to check merge with old value if needed
						const BoundType& child_bind = bindings->get( new_child );
						if ( child_bind.type ) {
							// store this insert under the original key, which must be a REM 
							// binding change if present
							auto bit = bedits.find( key );
							if ( bit != bedits.end() ) {
								BEdit& be = bit->second;
								assertf( be.mode == Bindings::REM, "inconsistent binding" );
								be = BEdit{ Bindings::INS, child_bind };
							} else {
								bedits.emplace_hint( bit, key, BEdit{ Bindings::INS, child_bind } );
							}
						}
						// remove child binding from map
						bindings = bindings->erase( new_child );
					}
				} break;
				case Classes::REMFROM: {
					// do not split classes
				} break;
				default: assertf( false, "invalid mode" );
			}
		}

		// finish merging bindings. This may fail, or may merge further classes
		for ( const auto& entry : bedits ) {
			interned_string key = entry.first;
			const BEdit& e = entry.second;

			// skip binding removals; the ADDTO merge has taken out all the new non-roots
			if ( e.mode == Bindings::REM ) continue;
			assertf( e.mode == Bindings::UPD || e.mode == Bindings::INS, "inconsistent binding" );

			// attempt binding update
			interned_string root = classes->find( key );
			BoundType ebound = bindings->get( root );
			const BoundType& nbound = e.val;

			if ( ebound.type && nbound.type ) {
				// attempt to unify bound classes
				Type* common = nullptr;
				AssertionSet need, have;
				OpenVarSet openVars;
				if ( unifyInexact( ebound.type->clone(), nbound.type->clone(), *this, need, have, 
						openVars, WidenMode{ ebound.allowWidening, nbound.allowWidening }, 
						indexer, common ) ) {
					// update bindings
					ebound.allowWidening &= nbound.allowWidening;
					if ( common ) {
						common->get_qualifiers() = Type::Qualifiers{};
						ebound.type = common;
					}
					// reset root in case updated in course of unification
					bindings = bindings->set( classes->find( root ), ebound );
				} else return false;  // cannot unify bindings
			}
		}

		return true;
	}

	void TypeEnvironment::extractOpenVars( OpenVarSet &openVars ) const {
		bindings->for_each( [&]( interned_string root, const BoundType& bound ) {
			classes->for_class( root, [&openVars,&bound]( interned_string var ) {
				openVars[ var ] = bound.data;
			} );
		} );
	}

	void TypeEnvironment::addActual( const TypeEnvironment& actualEnv, OpenVarSet& openVars ) {
		PRE_POST_VALIDATE
		if ( classes == actualEnv.classes && bindings == actualEnv.bindings ) {
			// actualEnv already the same, just need to update openVars
			extractOpenVars( openVars );
			return;
		} else assertf(classes != actualEnv.classes && bindings != actualEnv.bindings, "classes & bindings updated together");

		// read out equivalence classes (avoids conflicting reroots)
		std::vector<EqvClass> ecs;
		actualEnv.bindings->for_each( [&]( interned_string root, const BoundType& bound ) {
			ecs.emplace_back(bound);
			actualEnv.classes->for_class( root, [&]( interned_string var ) {
				ecs.back().vars.push_back( var );
			} );
		} );

		// add members of new class, setting openVars concurrently
		for ( EqvClass& ec : ecs ) {
			// merge vars
			interned_string new_root{nullptr};
			for ( interned_string var : ec.vars ) {
				openVars[ var ] = ec.bound.data;
				Classes* new_classes = classes->add( var );
				if ( new_classes == classes ) {
					// xxx - this case is a bit sketchy, but has been working
					// xxx - merging the classes is a departure from previous behaviour
					var = classes->find( var );
				} else { classes = new_classes; }
				new_root = new_root ? mergeClasses( new_root, var ).first : var;
			}
			// set binding without widening
			bindings = bindings->set( new_root, 
				BoundType{ maybeClone(ec.bound.type), false, ec.bound.data } );
		}
	}

	void TypeEnvironment::forbidWidening() {
		PRE_POST_VALIDATE
		bindings = bindings->mutate_each([]( const interned_string&, BoundType& c ) {
			if ( c.allowWidening ) {
				option<BoundType> ret = c;
				c.allowWidening = false;
				return ret;
			} else return option<BoundType>{};
		});
	}

	std::ostream & operator<<( std::ostream & out, const TypeEnvironment & env ) {
		env.print( out );
		return out;
	}

	PassVisitor<GcTracer> & operator<<( PassVisitor<GcTracer> & gc, const TypeEnvironment & env ) {
		gc.pass.visit( env );
		// for ( ClassRef c : env ) {
		// 	maybeAccept( c.get_bound().type, gc );
		// }
		return gc;
	}
} // namespace ResolvExpr

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
